<?php
/**
 * @package         Captcha free
 * @version         24.4.0
 * 
 * @author          Logical-Arts <info@logical-arts.de>
 * @copyright       Copyright © 2023 Logical-Arts All Rights Reserved
 * @license         GNU GPLv3 <http://www.gnu.org/licenses/gpl.html> or later
 * @link            https://www.logical-arts-extensions.com
 */

namespace Larts\Plugin\Captcha\LartsCaptcha\Extension;
// phpcs:disable PSR1.Files.SideEffects
\defined('_JEXEC') or die;
// phpcs:enable PSR1.Files.SideEffects

use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\Utilities\IpHelper;
use Joomla\Event\SubscriberInterface;
use Joomla\Event\DispatcherInterface;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Event\Plugin\AjaxEvent;
use Joomla\CMS\Uri\Uri;

use Larts\Plugin\Captcha\LartsCaptcha\lib\Base\Logger;
use Larts\Plugin\Captcha\LartsCaptcha\lib\ImageGenerator;
use Larts\Plugin\Captcha\LartsCaptcha\lib\LartsPluginHelper;


class LartsCaptcha extends CMSPlugin implements SubscriberInterface
{
	const WAIT_HINT_NONE = 0;
	const WAIT_HINT_DISABLE_SUBMIT = 1;
	const WAIT_HINT_COUNTDOWN = 2;

	const IMGTYPE_NONE = 0;
	const IMGTYPE_CALC = 1;
	const IMGTYPE_TEXT = 2;
	const IMGTYPE_SELECT = 3;
	const IMGTYPE_RANDOM = 4;

	const DATA_FIELDID = 0;
	const DATA_HONEYID = 1;
	const DATA_TIME_TRESHOLD = 2;
	const DATA_IMGTYPE = 3;
	const DATA_RESULT = 4;
	
	protected $autoloadLanguage = true;		// Load the language file on instantiation.
	private $_skip;
	private $wait_enabled = false;
	private $wait_secs = 15;
	private $wait_hint = 0;
	private $honeypot_enabled = false;
	private $honey_id = null;
	private $imgtype = 0;
	private $use_blacklist = false;
	private $blacklist = '';
	private $logging_enabled = false;

	private $config_ok = false;
	private $script = '';
	private $imgdata = null;
	public $img_check = null;
	public $data = null;


	public function __construct(DispatcherInterface $dispatcher, array $config = [])
  {
		$this->setApplication(Factory::getApplication());

		parent::__construct($dispatcher, $config);

		if(empty($config)) {
			return;
		}

		$this->wait_enabled = $this->params->get('wait_enabled', 0);
		$this->honeypot_enabled = $this->params->get('honeypot_enabled', 1);
		$this->imgtype = $this->params->get('imgtype', 0);
		$this->logging_enabled = $this->params->get('logging_enabled', 0);


		$this->config_ok = ( $this->wait_enabled || $this->honeypot_enabled || $this->use_blacklist );

		$user = $this->getApplication()->getIdentity();

		// skip if logged in or accessing backend
		if($user->id != 0 || $this->getApplication()->isClient('administrator')) {
			$this->_skip = true;
		}
		
    // Some browsers will request the page multiple times to collect other data.
		if(!LartsPluginHelper::serverFetchDoc()) {
      $this->_skip = true;
		}
		
	}

	public static function getSubscribedEvents(): array
	{
		return [
			'onInit' => 'onInit',
			'onDisplay' => 'onDisplay',
			'onCheckAnswer' => 'onCheckAnswer',
		];
	}


	// Initialise the captcha
	public function onInit($name = 'jform_captcha')
	{
		// J5 event call?
		if(is_a($name, 'Joomla\Event\Event', false)) {
			$name = $name->getName();
		}
		
		$wa = $this->getApplication()->getDocument()->getWebAssetManager();
		$wa->registerAndUseStyle('lartscaptcha_css', 'plg_captcha_lartscaptcha/lartscaptcha.css');
		
		if($this->_skip) {
			return;
		}

		$id = $name;
		$doc = Factory::getDocument();

		$this->data = [ self::DATA_FIELDID => '',
										self::DATA_HONEYID => '',
										self::DATA_TIME_TRESHOLD => 0,
										self::DATA_IMGTYPE => self::IMGTYPE_NONE,
										self::DATA_RESULT => 0,
									];
		if ($this->wait_enabled) {
			$this->data[self::DATA_TIME_TRESHOLD] = time() + $this->wait_secs;

			if($this->wait_hint != self::WAIT_HINT_NONE) {
				$wa->registerAndUseScript('lartscaptcha_hint', 'plg_captcha_lartscaptcha/hint.js');
			}
		} 
		
		$hide_elements = array('#' . $id, '#' . $id . '-lbl', '#' . $id . ' + span.optional', '#' . $id . '-lbl + span.optional');

		if ($this->honeypot_enabled)
		{
			$a = ord('a');
			$z = ord('0');
			$honey_id = 'larts_' .chr($a + rand(0, 25)).chr($a + rand(0, 25)).chr($a + rand(0, 25)) .chr($z + rand(0, 9)).chr($z + rand(0, 9));
			
			$this->data[self::DATA_FIELDID] = $id;
			$this->data[self::DATA_HONEYID] = $honey_id;

			$hide_elements[] = '#' . $this->honey_id;
		}
		
		if ($this->imgtype == self::IMGTYPE_RANDOM) {
			$this->imgtype = rand(self::IMGTYPE_CALC, self::IMGTYPE_SELECT);
		}

		if ($this->imgtype == self::IMGTYPE_CALC) {
			$this->prepareImgCalc();
		}


		$doc->addStyleDeclaration(join(',', $hide_elements) . "{display: none !important;}");
		return true;
	}

	public function prepareImgTxt()
	{
		$words = [ 'welcome', 'hope', 'enjoy', 'free', 'space', 'nice',
							 'moon', 'spice', 'golum', 'cesar' ];
		$txt = $this->params->get('imgtxtwords', false);
    if($txt !== false) {
			$words = explode(',', $txt);
		}

		$word = $words[mt_rand(0, count($words)-1)];

		$imggen = new ImageGenerator;
		$imggen->setPath(JPATH_PLUGINS. '/captcha/lartscaptcha/img/gen/');
		$imgpath = $imggen->generate($word, 230, 70);
		$pinfo = pathinfo($imgpath);
		$this->img_check = [ $pinfo['basename'], $word ];
		$this->data[self::DATA_RESULT] = $this->img_check[1];
		$this->data[self::DATA_IMGTYPE] = self::IMGTYPE_TEXT;
	}
	
	public function prepareImgCalc()
	{
		$num1 = mt_rand(0, 20);
		$num2 = mt_rand(0, 20);
		if(mt_rand(0, 1) == 0) {
			$op = '+';
			$result = $num1 + $num2;
		}
		else {
			$op = '-';
			$result = $num1 - $num2;
		}

		$imggen = new ImageGenerator;
		$imggen->setPath(JPATH_PLUGINS. '/captcha/lartscaptcha/img/gen/');
		$imgpath = $imggen->generate(' '.$num1.$op.$num2, 230, 70);
		$pinfo = pathinfo($imgpath);
		
		$this->img_check = [ $pinfo['basename'], $result ];
		$this->data[self::DATA_RESULT] = $this->img_check[1];
		$this->data[self::DATA_IMGTYPE] = self::IMGTYPE_CALC;
	}
	
	// Show the captcha
	public function onDisplay($name = null, $id = 'captchaguard', $class = '')
	{
		$app = $this->getApplication();

		if($this->_skip) {
			return '';
		}

		if (!$this->config_ok) {
			return '<strong style="color:red">Captcha-Config-Error</strong>';
		}
		
		$captstr = implode('|', $this->data);
		$captstr = base64_encode($captstr);
		$html = '<input type="hidden" id="captinfo" name="captinfo" value="'.$captstr.'" />';

		$class = preg_replace('/\brequired(?:="required")?\b/i', '', trim($class));

		if ($this->wait_enabled) {
			if($this->wait_hint == self::WAIT_HINT_DISABLE_SUBMIT) {
				$html .= '<script>window.addEventListener("DOMContentLoaded", function(event) {'
							 . '	hint_disable('.$this->wait_secs.', "hinttime"); });</script>';
				$html .= '<div class="hinttime" id="hinttime"></div>';
			}
			else if($this->wait_hint == self::WAIT_HINT_COUNTDOWN) {
				$html .= '<script>window.addEventListener("DOMContentLoaded", function(event) {'
							 . '	hint_timer('.$this->wait_secs.', "hinttime"); });</script>';
				$html .= '<div class="hinttime" id="hinttime"></div>';
			}
		}
		
		if(!empty($this->script)) {
			$html .= '<script>'.$this->script.'</script>';
		}

		
		if ($this->honeypot_enabled) {
			$html .= sprintf('<input name="%s" id="%s" class="captcha-hide" %s />', $id, $id, trim($class)) ;
			$html .= sprintf('<input type="text" id="%s" name="%s" class="captcha-hide" />', $this->honey_id, $this->honey_id);
		}

		if ($this->imgtype == self::IMGTYPE_CALC) {
			$img = $this->img_check[0];
			
			// Get the path for layout file
			$path = PluginHelper::getLayoutPath('captcha', 'lartscaptcha', 'imgcalc');
			// Render the layout
			ob_start();
			include $path;
			$html_imgs = ob_get_clean();
			$html .= $html_imgs;
		}

		
		return $html;
	}

	// Check the captcha
	public function onCheckAnswer($code = null)
	{
		$app = $this->getApplication();

		if($this->_skip) {
			return true;
		}

		$remoteip = IpHelper::getIp();
		$input = $app->getInput();
		
		$captstr = $input->get('captinfo', '');
		if(empty($captstr)) {
			return true;
		}
		$captstr = base64_decode($captstr);
		$this->data = explode('|', $captstr);

	
		// Check for IP
		if (empty($remoteip)) {
			throw new \RuntimeException(Text::_('PLG_CAPTCHA_LARTSCAPTCHA_ERROR_NO_IP'));
		}

		if (!$this->config_ok) {
			$this->logReject('invalid plugin configuration');
			return false;
		} 
		if ($this->wait_enabled) {
			$now = time();
			$until = (int)(isset($this->data[self::DATA_TIME_TRESHOLD]) ? $this->data[self::DATA_TIME_TRESHOLD] : $now);
			if ($now <= $until) {
				$this->logReject('time threshold not reached (' . ( $until - $now ) . 's)');
				$this->set_error(Text::_('PLG_CAPTCHA_LARTSCAPTCHA_REJECTED'));
				return false;
			}
		}
		if ($this->honeypot_enabled) {
			$field_id = $this->data[self::DATA_FIELDID];
			if ($input->get($field_id, '', 'RAW') !== '') {
				$this->logReject('honeypot filled (trapped #' . $field_id . ')');
				$this->set_error(Text::_('PLG_CAPTCHA_LARTSCAPTCHA_REJECTED'));
				return false;
			}
			$honey_id = $this->data[self::DATA_HONEYID];
			if ($input->get($honey_id, '', 'RAW') !== '') {
				$this->logReject('honeypot filled (trapped #' . $honey_id . ')');
				$this->set_error(Text::_('PLG_CAPTCHA_LARTSCAPTCHA_REJECTED'));
				return false;
			}
		} 

		if ($this->imgtype == self::IMGTYPE_CALC) {
			$result = $this->data[self::DATA_RESULT];
			$user = $input->getInt('imgres', '');
			
			if((int)$user != (int)$result) {
				$this->logReject('invalid calc answer (' .$user. ' : ' .$result. ')');
				$this->set_error(Text::_('PLG_CAPTCHA_LARTSCAPTCHA_REJECTED'));
				return false;
			}
		}


		return true;
	}

	// Method to react on the setup of a captcha field. Gives the possibility 
	// to change the field and/or the XML element for the field.
	public function onSetupField(\Joomla\CMS\Form\Field\CaptchaField $field, \SimpleXMLElement $element)
	{
			// Hide the label for the invisible recaptcha type
			$element['hiddenLabel'] = 'true';
	}
	
	private function logReject($msg)
	{
		if (!$this->logging_enabled) {
			return;
		}
		Logger::enable(true);
		Logger::_('lartscaptcha', 'REJECT: '. $msg);
	}
	
	private function set_error($msg) {
		return $this->getApplication()->enqueueMessage($msg, 'error');
	}

	
}

